"""
Abstraction layer for questions. Each class represents different question types.

The questions describe literal questions that are asked of the user in an interactive
mode via the ``inquirer`` package.

Questions will be skipped and use the default value if the `ignore` property resolves
to True.

Questions will also resolve to their default values if running in "fast mode".
"""

from copy import deepcopy
from functools import partial
from typing import Any, Callable, Dict, Optional, Type, TypeVar, cast

import inquirer
import inquirer.errors
import inquirer.questions

from jobbergate_cli.exceptions import Abort
from jobbergate_cli.render import render_dict
from jobbergate_cli.subapps.applications.application_base import JobbergateApplicationBase


TInquirerType = TypeVar("TInquirerType", bound=inquirer.questions.Question)


class QuestionBase:
    """
    Baseclass for questions.

    All questions have variablename, message and an optional default.
    """

    def __init__(
        self,
        variablename: str,
        message: str,
        ignore: bool = False,
        default: Optional[Any] = None,
        inquirer_type: Type[TInquirerType] = inquirer.Text,
    ):
        """
        Initialize the Question.

        :param: variablename:  The key in the config dictionary that this question will set
        :param: message:       The message to show the user that describes what the question is gathering
        :param: ignore:        If true, do not ask the question and just use the default value instead
        :param: default:       The default value for the variablename in the answers dict
        :param: inquirer_type: The ``inquirer`` question type that this ``QuestionBase`` wraps
        """
        self.variablename = variablename
        self.default = default
        self.inquirer_kwargs = dict(
            message=message,
            default=default,
            ignore=ignore,
        )
        self.inquirer_type = inquirer_type

    def make_prompts(self, **override_kwargs):
        """
        Create ``inquirer`` prompts from this instance of ``QuestionBase``.

        :param: override_kwargs: A collection of keyword arguments to override in intializing the ``inquirer`` question
        """
        final_kwargs = {
            **self.inquirer_kwargs,
            **override_kwargs,
        }
        return [self.inquirer_type(self.variablename, **final_kwargs)]


class Text(QuestionBase):
    """
    Asks for a text value.
    """


class Integer(QuestionBase):
    """
    Asks for an integer value. Could have min and/or max constrains.
    """

    def __init__(
        self,
        variablename: str,
        message: str,
        minval: Optional[int] = None,
        maxval: Optional[int] = None,
        **kwargs,
    ):
        """
        Initialize the Integer question.

        :param: variablename:  The key in the config dictionary that this question will set
        :param: message:       The message to show the user that describes what the question is gathering
        :param: minval:        The minimum value the integer may be set to. If not specified, use negative infinity.
        :param: minval:        The maximum value the integer may be set to. If not specified, use infinity.
        """
        super().__init__(variablename, message, **kwargs)
        self.minval = minval
        self.maxval = maxval
        self.inquirer_kwargs.update(validate=self._validator)

    def _validator(self, _, current):
        """
        Provide a custom validator that checks the value of the integer to make sure it is in range.
        """
        try:
            int_val = int(current)
        except ValueError:
            raise inquirer.errors.ValidationError("", reason=f"{current} is not an integer")

        min_str = str(self.minval) if self.minval is not None else "-∞"
        max_str = str(self.maxval) if self.maxval is not None else "∞"
        if any(
            [
                self.minval is not None and int_val < self.minval,
                self.maxval is not None and int_val > self.maxval,
            ]
        ):
            raise inquirer.errors.ValidationError("", reason=f"{current} is out of range [{min_str}, {max_str}]")
        return True


class List(QuestionBase):
    """
    Gives the user a list to choose one from.
    """

    def __init__(self, variablename: str, message: str, choices: list, **kwargs):
        """
        Initialize the List question.

        :param: variablename:  The key in the config dictionary that this question will set
        :param: message:       The message to show the user that describes what the question is gathering
        :param: choices:       A list of the possible values from which the Question will allow the user to select one
        """
        super().__init__(variablename, message, inquirer_type=inquirer.List, **kwargs)
        self.inquirer_kwargs.update(choices=choices)


class Directory(QuestionBase):
    """
    Asks for a directory name. If `exists` is `True` it checks if path exists and is a directory.

    :param exists: Checks if given directory exists
    """

    def __init__(self, variablename: str, message: str, exists: Optional[bool] = None, **kwargs):
        """
        Initialize the Directory question.

        :param: variablename:  The key in the config dictionary that this question will set
        :param: message:       The message to show the user that describes what the question is gathering
        :param: exists:        If True, ensure that the directory exists on the system
        """

        super().__init__(variablename, message, inquirer_type=inquirer.Path, **kwargs)
        self.inquirer_kwargs.update(path_type=inquirer.Path.DIRECTORY)
        if exists is not None:
            self.inquirer_kwargs.update(exists=exists)


class File(QuestionBase):
    """
    Asks for a file name.
    """

    def __init__(self, variablename: str, message: str, exists: Optional[bool] = None, **kwargs):
        """
        Initialize the File question.

        :param: variablename:  The key in the config dictionary that this question will set
        :param: message:       The message to show the user that describes what the question is gathering
        :param: exists:        If True, ensure that the file path exists on the system
        """
        super().__init__(variablename, message, inquirer_type=inquirer.Path, **kwargs)
        self.inquirer_kwargs.update(path_type=inquirer.Path.FILE)
        if exists is not None:
            self.inquirer_kwargs.update(exists=exists)


class Checkbox(QuestionBase):
    """
    Gives the user a list to choose multiple entries from.
    """

    def __init__(self, variablename: str, message: str, choices: list, **kwargs):
        """
        Initialize the Checkbox question.

        :param: variablename:  The key in the config dictionary that this question will set
        :param: message:       The message to show the user that describes what the question is gathering
        :param: choices:       A list of the possible values from which the Question will allow the user to select many
        """

        super().__init__(
            variablename,
            message + " (press CTRL+A to SELECT ALL or CTRL+R to DESELECT ALL)",
            inquirer_type=inquirer.Checkbox,
            **kwargs,
        )
        self.inquirer_kwargs.update(choices=choices)


class Confirm(QuestionBase):
    """
    Asks a question with a boolean answer (true/false).
    """

    def __init__(self, variablename: str, message: str, **kwargs):
        """
        Initialize the Confirm question.

        :param: variablename:  The key in the config dictionary that this question will set
        :param: message:       The message to show the user that describes what the question is gathering
        """
        super().__init__(variablename, message, inquirer_type=inquirer.Confirm, **kwargs)


class BooleanList(Confirm):
    """
    Asks a confirmation question that is followed up by a certain question list when true and a different list if false.
    """

    def __init__(
        self,
        variablename: str,
        message: str,
        whentrue=None,
        whenfalse=None,
        **kwargs,
    ):
        """
        Initialize the Checkbox question.

        :param: variablename:  The key in the config dictionary that this question will set
        :param: message:       The message to show the user that describes what the question is gathering
        :param whentrue:       List of questions to ask if user answers 'true' on this question
        :param whentrue:       List of questions to show if user answers 'false' on this question
        """
        super().__init__(variablename, message, **kwargs)
        self.whentrue_child_map = {c.variablename: c for c in (whentrue if whentrue is not None else [])}
        self.whenfalse_child_map = {c.variablename: c for c in (whenfalse if whenfalse is not None else [])}

    def ignore_child(self, child: QuestionBase, answers: Dict[str, Any]) -> bool:
        """
        Dynamically check if a child question should be ignored based on the questions that have already been answered.

        :param: child:   The child question that might be ignored
        :param: answers: Answer values to previously asked questions
        """
        my_answer = answers.get(self.variablename)
        Abort.require_condition(
            my_answer is not None,
            "Questions were asked out of order. Please check your Application for consistency",
        )
        if my_answer is True:
            if child.variablename in self.whentrue_child_map:
                return False
            elif child.variablename in self.whenfalse_child_map:
                return True
            else:
                return False  # This child wasn't registered. This should not happen. But, don't ignore to be safe.
        else:
            if child.variablename in self.whentrue_child_map:
                return True
            elif child.variablename in self.whenfalse_child_map:
                return False
            else:
                return False  # This child wasn't registered. This should not happen. But, don't ignore to be safe.

    def make_ignore_partial(self, child: QuestionBase) -> Callable[[Dict[str, Any]], bool]:
        """
        Build a partial method for checking if a child should be ignored.

        This method just makes the code more readable so that a non-descriptive lambda does not need to be used inline.
        """
        return partial(self.ignore_child, child)

    def make_prompts(self, **override_kwargs):
        """
        Create ``inquirer`` prompts from this instance of ``BooleanList`` and for all its child questions.

        :param: override_kwargs: A collection of keyword arguments to override in the base ``make_prompts`` method
        """

        retval = super().make_prompts(**override_kwargs)
        all_children = [*self.whentrue_child_map.values(), *self.whenfalse_child_map.values()]
        for child in all_children:
            retval.extend(child.make_prompts(ignore=self.make_ignore_partial(child)))
        return retval


class Const(Text):
    """
    Sets the variable to the `default` value. Doesn't show anything.
    """

    def __init__(self, variablename: str, **kwargs):
        """
        Initialize the Const "question".

        :param: variablename:  The key in the config dictionary that this question will set
        """
        super().__init__(variablename, "", **kwargs)

    def make_prompts(self):
        """
        Create ``inquirer`` prompts from this instance of ``Const``.
        """
        return super().make_prompts(ignore=True)


def gather_param_values(
    application: JobbergateApplicationBase,
    supplied_params: Optional[Dict[str, Any]] = None,
    fast_mode: bool = False,
) -> Dict[str, Any]:
    """
    Gather the parameter values by executing the application methods.

    Prompt users for answers or use defaults as needed.

    :param: application:     The application instance to pull questions from
    :param: supplied_params: Pre-supplied parameters.
                             Any questions where the variablename matches a pre-supplied key in the dict
                             at the start of execution will be skipped.
    :param: fast_mode:       Do not ask the user questions. Just use the supplied params and defaults.
    :returns: A dict of the gathered parameter values
    """
    if supplied_params is None:
        supplied_params = dict()

    config = deepcopy(supplied_params)

    next_method = "mainflow"

    while next_method is not None:
        method_to_call = getattr(application, next_method)

        try:
            workflow_questions = method_to_call(data=config)
        except NotImplementedError:
            raise Abort(
                f"""
                Abstract method not implemented.

                Please implement {method_to_call.__name__} in your class.",
                """,
                subject="Invalid application module",
            )

        prompts = []
        auto_answers = {}

        for question in workflow_questions:
            if question.variablename in supplied_params:
                continue
            elif fast_mode and question.default is not None:
                auto_answers[question.variablename] = question.default
            else:
                prompts.extend(question.make_prompts())

        workflow_answers = cast(Dict[str, Any], inquirer.prompt(prompts, raise_keyboard_interrupt=True))
        config.update(workflow_answers)
        config.update(auto_answers)
        if len(auto_answers) > 0:
            render_dict(auto_answers, title="Default values used")

        next_method = config.pop("nextworkflow", None)

    return config



"""
Provide constants that may be used throughout the CLI modules.
"""

from enum import Enum
from pathlib import Path


JOBBERGATE_APPLICATION_CONFIG = {
    "application_name": "",
    "application_description": "",
}

JOBBERGATE_JOB_SCRIPT_CONFIG = {
    "job_script_name": "",
    "job_script_description": "TEST_DESC",
    "job_script_owner_email": "",
    "application_id": "",
}

JOBBERGATE_JOB_SUBMISSION_CONFIG = {
    "job_submission_name": "",
    "job_submission_description": "TEST_DESC",
    "job_submission_owner_email": "",
    "job_script_id": "",
}

JOBBERGATE_DEFAULT_DOTENV_PATH = Path("/etc/default/jobbergate3-cli")
JOBBERGATE_APPLICATION_SUPPORTED_FILES = {".py", ".yaml", ".j2", ".jinja2"}
JOBBERGATE_APPLICATION_MODULE_FILE_NAME = "jobbergate.py"
JOBBERGATE_APPLICATION_CONFIG_FILE_NAME = "jobbergate.yaml"
TAR_NAME = "jobbergate.tar.gz"

OV_CONTACT = "Omnivector Solutions <info@omnivector.solutions>"


class SortOrder(str, Enum):
    """
    Enum descring the type of sort orders that are available for list commands.
    """

    ASCENDING = "ASCENDING"
    DESCENDING = "DESCENDING"
    UNSORTED = "UNSORTED"


class FileType(str, Enum):
    """File type enum."""

    ENTRYPOINT = "ENTRYPOINT"
    SUPPORT = "SUPPORT"



"""
Utilities for handling auth in jobbergate-cli.
"""

import webbrowser
from time import sleep
from typing import Dict, Optional, cast

from jose import jwt
from jose.exceptions import ExpiredSignatureError
from loguru import logger
from pydantic import ValidationError
from rich.console import Console

from jobbergate_cli.config import settings
from jobbergate_cli.exceptions import Abort, JobbergateCliError
from jobbergate_cli.render import terminal_message
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import DeviceCodeData, IdentityData, JobbergateContext, Persona, TokenSet
from jobbergate_cli.text_tools import copy_to_clipboard, unwrap
from jobbergate_cli.time_loop import TimeLoop


def validate_token_and_extract_identity(token_set: TokenSet) -> IdentityData:
    """
    Validate the access_token from a TokenSet and extract the user's identity data.

    Validations:
        * Checks if access_token is not empty.
        * Checks timestamp on the access token.
        * Checks that the client_id is present
        * Checks that email is present

    Reports an error in the logs and to the user if there is an issue with the access_token.
    """
    logger.debug("Validating access token")

    token_file_is_empty = not token_set.access_token
    if token_file_is_empty:
        logger.debug("Access token file exists but it is empty")
        raise Abort(
            """
            Access token file exists but it is empty.

            Please try logging in again.
            """,
            subject="Empty access token file",
            support=True,
            log_message="Empty access token file",
            sentry_context=dict(access_token=dict(access_token=token_set.access_token)),
        )

    try:
        token_data = jwt.decode(
            token_set.access_token,
            None,
            options=dict(
                verify_signature=False,
                verify_aud=False,
                verify_exp=True,
            ),
        )
    except ExpiredSignatureError:
        raise  # Will be handled in calling context
    except Exception as err:
        raise Abort(
            """
            There was an unknown error while validating the access token.

            Please try logging in again.
            """,
            subject="Invalid access token",
            support=True,
            log_message=f"Unknown error while validating access access token: {err}",
            sentry_context=dict(access_token=dict(access_token=token_set.access_token)),
            original_error=err,
        )

    logger.debug("Extracting identity data from the access token")
    try:
        identity = IdentityData(
            email=token_data.get("email"),
            client_id=token_data.get("azp"),
        )
    except ValidationError as err:
        raise Abort(
            """
            There was an error extracting the user's identity from the access token.

            Please try logging in again.
            """,
            subject="Missing user data",
            support=True,
            log_message=f"Token data could not be extracted to identity: {err}",
            original_error=err,
        )
    return identity


def load_tokens_from_cache() -> TokenSet:
    """
    Loads an access token (and a refresh token if one exists) from the cache.
    """

    # Make static type checkers happy
    assert settings.JOBBERGATE_ACCESS_TOKEN_PATH is not None
    assert settings.JOBBERGATE_REFRESH_TOKEN_PATH is not None

    Abort.require_condition(
        settings.JOBBERGATE_ACCESS_TOKEN_PATH.exists(),
        "Please login with your auth token first using the `jobbergate login` command",
        raise_kwargs=dict(subject="You need to login"),
    )

    logger.debug("Retrieving access token from cache")
    token_set = TokenSet(access_token=settings.JOBBERGATE_ACCESS_TOKEN_PATH.read_text())

    if settings.JOBBERGATE_REFRESH_TOKEN_PATH.exists():
        logger.debug("Retrieving refresh token from cache")
        token_set.refresh_token = settings.JOBBERGATE_REFRESH_TOKEN_PATH.read_text()

    return token_set


def save_tokens_to_cache(token_set: TokenSet):
    """
    Saves tokens from a token_set to the cache.
    """

    # Make static type checkers happy
    assert settings.JOBBERGATE_ACCESS_TOKEN_PATH is not None
    assert settings.JOBBERGATE_REFRESH_TOKEN_PATH is not None

    logger.debug(f"Caching access token at {settings.JOBBERGATE_ACCESS_TOKEN_PATH}")
    settings.JOBBERGATE_ACCESS_TOKEN_PATH.write_text(token_set.access_token)
    settings.JOBBERGATE_ACCESS_TOKEN_PATH.chmod(0o600)

    if token_set.refresh_token is not None:
        logger.debug(f"Caching refresh token at {settings.JOBBERGATE_REFRESH_TOKEN_PATH}")
        settings.JOBBERGATE_REFRESH_TOKEN_PATH.write_text(token_set.refresh_token)
        settings.JOBBERGATE_REFRESH_TOKEN_PATH.chmod(0o600)


def clear_token_cache():
    """
    Clears the token cache.
    """
    logger.debug("Clearing cached tokens")

    logger.debug(f"Removing access token at {settings.JOBBERGATE_ACCESS_TOKEN_PATH}")
    if settings.JOBBERGATE_ACCESS_TOKEN_PATH.exists():
        settings.JOBBERGATE_ACCESS_TOKEN_PATH.unlink()

    logger.debug(f"Removing refresh token at {settings.JOBBERGATE_REFRESH_TOKEN_PATH}")
    if settings.JOBBERGATE_REFRESH_TOKEN_PATH.exists():
        settings.JOBBERGATE_REFRESH_TOKEN_PATH.unlink()


def init_persona(ctx: JobbergateContext, token_set: Optional[TokenSet] = None):
    """
    Initializes the "persona" which contains the tokens and email address for a user.

    Retrieves the access token for the user from the cache.

    Token is retrieved from the cache, validated, and user email is extracted.

    If the access token is expired, a new one will be acquired via the cached refresh token (if there is one).

    Saves token_set to cache.

    Returns the persona.
    """
    if token_set is None:
        token_set = load_tokens_from_cache()

    try:
        identity_data = validate_token_and_extract_identity(token_set)
    except ExpiredSignatureError:
        Abort.require_condition(
            token_set.refresh_token is not None,
            "The auth token is expired. Please retrieve a new and log in again.",
            raise_kwargs=dict(
                subject="Expired access token",
                support=True,
            ),
        )

        logger.debug("The access token is expired. Attempting to refresh token")
        refresh_access_token(ctx, token_set)
        identity_data = validate_token_and_extract_identity(token_set)

    logger.debug(f"Persona created with identity_data: {identity_data}")

    save_tokens_to_cache(token_set)

    return Persona(
        token_set=token_set,
        identity_data=identity_data,
    )


def refresh_access_token(ctx: JobbergateContext, token_set: TokenSet):
    """
    Attempt to fetch a new access token given a refresh token in a token_set.

    Sets the access token in-place.

    If refresh fails, notify the user that they need to log in again.
    """
    url = f"https://{settings.OIDC_DOMAIN}/protocol/openid-connect/token"
    logger.debug(f"Requesting refreshed access token from {url}")

    JobbergateCliError.require_condition(
        ctx.client is not None,
        "Attempted to refresh with a null client. This should not happen",
    )

    # Make static type-checkers happy
    assert ctx.client is not None

    refreshed_token_set = cast(
        TokenSet,
        make_request(
            # Can this even work? this client should be for the armada api...
            ctx.client,
            "/protocol/openid-connect/token",
            "POST",
            abort_message="The auth token could not be refreshed. Please try logging in again.",
            abort_subject="EXPIRED ACCESS TOKEN",
            support=True,
            response_model_cls=TokenSet,
            data=dict(
                client_id=settings.OIDC_CLIENT_ID,
                audience=settings.OIDC_AUDIENCE,
                grant_type="refresh_token",
                refresh_token=token_set.refresh_token,
            ),
        ),
    )

    token_set.access_token = refreshed_token_set.access_token
    if refreshed_token_set.refresh_token is not None:
        token_set.refresh_token = refreshed_token_set.refresh_token


def open_on_browser(url: str) -> bool:
    """Open the url on the browser using webbrowser."""
    try:
        browser = webbrowser.get()
        browser.open(url)
        return True
    except Exception as e:
        logger.warning(f"Couldn't open login url on browser due to -- {str(e)}")
        return False


def show_login_message(verification_uri: str):
    """Show a message to the user with a link to the auth provider to login."""
    console = Console()
    EXTRA_CHARS = 7  # for indentation and panel borders

    kwargs = {}

    if open_on_browser(verification_uri):
        kwargs["footer"] = "The output was opened on your browser"
    elif copy_to_clipboard(verification_uri):
        kwargs["footer"] = "The output was copied to your clipboard"

    if console.width >= len(verification_uri) + EXTRA_CHARS:
        _show_login_standard_screen(verification_uri, **kwargs)
    else:
        _show_login_narrow_screen(verification_uri, console, **kwargs)


def _show_login_narrow_screen(verification_uri: str, console: Console, **kwargs):
    """Print the link out of the panel to make it easier to copy."""
    terminal_message(
        f"""
        To complete login, please open the link bellow in a browser.

        Waiting up to {settings.OIDC_MAX_POLL_TIME / 60} minutes for you to complete the process...
        """,
        subject="Waiting for login",
        **kwargs,
    )
    console.print(verification_uri, overflow="ignore", no_wrap=True, crop=False)
    console.print()


def _show_login_standard_screen(verification_uri: str, **kwargs):
    """Print a rich panel with a link to the auth provider to login."""
    terminal_message(
        f"""
        To complete login, please open the following link in a browser:

          {verification_uri}

        Waiting up to {settings.OIDC_MAX_POLL_TIME / 60} minutes for you to complete the process...
        """,
        subject="Waiting for login",
        **kwargs,
    )


def fetch_auth_tokens(ctx: JobbergateContext) -> TokenSet:
    """
    Fetch an access token (and possibly a refresh token) from Auth0.

    Prints out a URL for the user to use to authenticate and polls the token endpoint to fetch it when
    the browser-based process finishes
    """
    # Make static type-checkers happy
    assert ctx.client is not None

    device_code_data = cast(
        DeviceCodeData,
        make_request(
            ctx.client,
            "/protocol/openid-connect/auth/device",
            "POST",
            expected_status=200,
            abort_message="There was a problem retrieving a device verification code from the auth provider",
            abort_subject="COULD NOT RETRIEVE TOKEN",
            support=True,
            response_model_cls=DeviceCodeData,
            data=dict(
                client_id=settings.OIDC_CLIENT_ID,
                grant_type="client_credentials",
                audience=settings.OIDC_AUDIENCE,
            ),
        ),
    )
    show_login_message(device_code_data.verification_uri_complete)

    for tick in TimeLoop(
        settings.OIDC_MAX_POLL_TIME,
        message="Waiting for web login",
    ):
        response_data = cast(
            Dict,
            make_request(
                ctx.client,
                "/protocol/openid-connect/token",
                "POST",
                abort_message="There was a problem retrieving a device verification code from the auth provider",
                abort_subject="COULD NOT FETCH ACCESS TOKEN",
                support=True,
                data=dict(
                    grant_type="urn:ietf:params:oauth:grant-type:device_code",
                    device_code=device_code_data.device_code,
                    client_id=settings.OIDC_CLIENT_ID,
                ),
            ),
        )
        if "error" in response_data:
            if response_data["error"] == "authorization_pending":
                logger.debug(f"Token fetch attempt #{tick.counter} failed")
                sleep(device_code_data.interval)
            else:
                # TODO: Test this failure condition
                raise Abort(
                    unwrap(
                        """
                        There was a problem retrieving a device verification code from the auth provider:
                        Unexpected failure retrieving access token.
                        """
                    ),
                    subject="Unexpected error",
                    support=True,
                    log_message=f"Unexpected error response: {response_data}",
                )
        else:
            return TokenSet(**response_data)

    raise Abort(
        "Login process was not completed in time. Please try again.",
        subject="Timed out",
        log_message="Timed out while waiting for user to complete login",
    )
